from collections.abc import Sized, Iterator
import logging
import subprocess
import sys
import traceback
from typing import Any, Callable, List, TypeVar, Generic, Iterable
from click import echo
import pandas as pd
from result import Err, Ok, Result


def sh(cmd: str, default: List[str] = []) -> Result[List[str], str]:
    process = subprocess.run(
        [arg for arg in cmd.split(" ")],
        capture_output=True,
        text=True,
    )
    match process.returncode:
        case 0:
            return Ok([li.strip() for li in process.stdout.splitlines()])
        case _:
            echo(process.stderr)
            if default:
                return Ok(default)
            else:
                return Err(process.stderr)


def proc(cmd: str, timeout: float | None = None) -> Result[List[str], str]:
    process = subprocess.Popen(
        [arg for arg in cmd.split(" ")],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
    )
    if process.stdout is not None:
        for line in iter(process.stdout.readline, ""):
            echo(line.strip())
    return_code = process.wait(timeout=timeout)
    match return_code:
        case 0:
            return Ok([])
        case _:
            if process.stderr is not None:
                return Err("\n".join(process.stderr.readlines()))
    return Err("failed!")


def set_global(name: str, value: Any):
    globals()[name] = value


def get_global(name: str, default: Any = None) -> Any:
    try:
        return globals()[name]
    except KeyError:
        return None if default is None else default


def incr(name: str) -> Any:
    value = get_global(name)
    if value is None:
        set_global(name, 0)
        return 0
    else:
        value += 1
        set_global(name, value)
        return value


class _logger:
    def __init__(self, name: str) -> None:
        set_global("_logger_name", name)
        self._hide_level_name = False

    def level(self, level: int):
        _get_logger().setLevel(level)
        return self

    def debug(self):
        return self.level(logging.DEBUG)

    def info(self):
        return self.level(logging.INFO)

    def warn(self):
        return self.level(logging.WARNING)

    def error(self):
        return self.level(logging.ERROR)

    def critical(self):
        return self.level(logging.CRITICAL)

    def use_print(self, use_print: bool):
        set_global("_use_print", use_print)
        return self

    def callback(self, callback: Callable[[str], None]):
        set_global("_callback", callback)
        return self

    def hide_level_name(self):
        self._hide_level_name = True

    def basic_config(self, level: int = logging.WARN):
        logging.basicConfig(
            format="%(asctime)s %(message)s"
            if self._hide_level_name
            else "%(asctime)s %(levelname)s %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
            # NB: this is the default level of the root logger
            level=level,
            stream=sys.stdout,
        )
        return self

    def fn_level(self, level: int):
        """
        Set the level of the `log(message: str)` function.
        """
        set_global("_log_fn_level", level)
        return self

    def fn_debug(self):
        return self.fn_level(logging.DEBUG)

    def fn_info(self):
        return self.fn_level(logging.INFO)

    def fn_warn(self):
        return self.fn_level(logging.WARNING)

    def fn_error(self):
        return self.fn_level(logging.ERROR)

    def fn_critical(self):
        return self.fn_level(logging.CRITICAL)


def logger(name: str) -> _logger:
    return _logger(name)


def __log_fn(level: int) -> Callable[[str], None]:
    logger = _get_logger()
    if level == logging.DEBUG:
        return logger.debug
    elif level == logging.INFO:
        return logger.info
    elif level == logging.WARNING:
        return logger.warning
    elif level == logging.ERROR:
        return logger.error
    elif level == logging.CRITICAL:
        return logger.critical
    else:
        return logging.info


def _if_level(level: int, message: str):
    _use_print = get_global("_use_print", False)
    _callback = get_global("_callback", None)
    _logger = _get_logger()

    if _logger.level <= level:
        if _use_print:
            print(f"{message}", flush=True)
        if _callback:
            _callback(message)


def _get_logger() -> logging.Logger:
    logger_name = get_global("_logger_name")
    return logging.getLogger(logger_name)


def log(message: Any):
    message = str(message)
    level: int = get_global("_log_fn_level", logging.INFO)
    _if_level(level, message)
    __log_fn(level)(message)


def dbg(message: Any):
    message = str(message)
    logger = _get_logger()
    _if_level(logging.DEBUG, message)
    logger.debug(message)


def debug(message: Any):
    message = str(message)
    logger = _get_logger()
    _if_level(logging.DEBUG, message)
    logger.debug(message)


def info(message: Any):
    message = str(message)
    logger = _get_logger()
    _if_level(logging.INFO, message)
    logger.info(message)


def warn(message: Any):
    message = str(message)
    logger = _get_logger()
    _if_level(logging.WARNING, message)
    logger.warning(message)


def err(message: Any):
    message = str(message)
    logger = _get_logger()
    _if_level(logging.ERROR, message)
    logger.error(message)


def tb():
    if sys.exc_info()[0]:  # type: ignore # noqa
        message = traceback.format_exc()
        logger = _get_logger()
        _if_level(logging.ERROR, message)
        logger.error(message)


A = TypeVar("A")


class progress(Generic[A]):
    def __init__(
        self,
        iterator: Iterable[A],
        name: str = "",
        num_steps: int = 10,
        precision: int = 1,
        print_idx: bool = False,
    ) -> None:
        if not isinstance(iterator, Iterable):  # type: ignore
            raise ValueError("Passed object is not iterable")
        if not isinstance(iterator, Sized):
            raise ValueError("Passed object does not have a size")

        self.print_idx = print_idx
        self.iterator: Iterator[A] = iter(iterator)
        if isinstance(iterator, pd.DataFrame):
            self.iterator = iterator.iterrows()  # type: ignore
        self.len = len(iterator)
        self.name = name
        self.num_steps = num_steps
        if self.num_steps > self.len:
            self.num_steps = self.len

        if self.num_steps < 1:
            self.num_steps = 1

        self.map = {}
        step_size = 100 / self.num_steps
        is_int = step_size.is_integer()
        for i in range(self.num_steps + 1):
            index = i * int(self.len / self.num_steps)
            percent = i * step_size
            percent_str = f"{percent:.{precision}f}%"
            if is_int:
                percent_str = f"{int(percent):d}%"
            self.map[index] = percent_str

        self.indices = set(self.map.keys())
        self.index = 0
        self.index_len = len(str(self.len))
        self.percent_len = max([len(p) for p in self.map.values()])

    def __iter__(self):
        return self

    def __next__(self) -> A:
        if self.index in self.indices:
            msg = f"{self.name} {self.map[self.index]:>{self.percent_len}}"
            if self.print_idx:
                msg += f" n={self.index:<{self.index_len}}"
            log(msg)
        if self.index >= self.len:
            raise StopIteration
        else:
            self.index += 1
            return next(self.iterator)


if __name__ == "__main__":
    print(get_global("asdf"))
    set_global("asdf", 0)
    incr("asdf")
    print(get_global("asdf"))

    print(get_global("zcxv"))
    incr("zcxv")
    print(get_global("zcxv"))

    # def rev(msg):
    #     print(msg[::-1])
    #     pass

    # import time

    # logger("test").info().basic_config()

    # # too lazy rn to make a proper test suite
    # for i in progress(range(101), name="test1"):
    #     time.sleep(0.001)

    # for i in progress(
    #     range(7),
    #     name="test2",
    #     print_idx=True,
    #     num_steps=11,
    #     precision=3,
    # ):
    #     time.sleep(0.001)

    # for idx, i in enumerate(progress(["a", "b", "c"], name="test3")):
    #     time.sleep(0.001)

    # for i in progress([0.1, 0.2, 0.3], name="test4"):
    #     log(i)
    #     time.sleep(0.001)

    # log("hello world")

    # # logging.getLogger("ayy").setLevel(logging.INFO)
    # logger("ayy").info().fn_info().basic_config()

    # set_global("ayy", "lmao")
    # value = get_global("ayy")
    # log(value)

    # from datetime import datetime

    # err(datetime.now())

    # for i in progress([], name="test5"):
    #     time.sleep(0.001)

    # a = pd.DataFrame([{"a": 1, "b": 2}, {"a": 3, "b": 4}])

    # for idx, row in progress(a, name="test6"):  # type: ignore
    #     log(row)
    #     time.sleep(0.001)



"""
simple key-value store using sqlite3

shamelessly stolen from https://pypi.org/project/kv/
now it's typed + we can add default values
"""

from contextlib import contextmanager
import sqlite3
from collections.abc import MutableMapping
from typing import Any
import json


class KV(MutableMapping[str, Any]):
    def __init__(
        self,
        db: str = ":memory:",
        table: str = "kv",
        default: Any = None,
        timeout: int = 5,
    ):
        self._db_uri = db
        self._table = table
        self._default = default
        self._db = sqlite3.connect(db, timeout=timeout)
        self._db.isolation_level = None
        self._attr: str = None  # type: ignore
        self._locks = 0
        self._execute(
            f"""
            CREATE TABLE IF NOT EXISTS {self._table} 
            (key PRIMARY KEY, value)
            """
        )

    def incr(self, key: str, amount: int = 1) -> int:
        with self.lock():
            val = self[key]
            val += amount
            self[key] = val
            return val

    def decr(self, key: str, amount: int = 1) -> int:
        with self.lock():
            val = self[key]
            val -= amount
            self[key] = val
            return val

    def clone(self, table: str) -> "KV":
        return KV(self._db_uri, table)

    def _execute(self, sql: str, params: tuple[str, ...] = ()):
        if params:
            return self._db.cursor().execute(sql, params)
        else:
            return self._db.cursor().execute(sql)

    def __len__(self):
        [[n]] = self._execute(f"SELECT COUNT(*) FROM {self._table}")
        return n

    def __getitem__(self, key: str) -> int:
        if key is None:  # type: ignore
            raise ValueError("key cannot be None")

        for row in self._execute(
            f"""
            SELECT value FROM {self._table} WHERE key=?
            """,
            (key,),
        ):
            return json.loads(row[0])
        else:
            if self._default is not None:
                return self._default
            else:
                raise KeyError

    def __iter__(self):
        return (
            key for [key] in self._execute(f"SELECT key FROM {self._table}")
        )

    def __setitem__(self, key: str, value: Any):
        jvalue = json.dumps(value)
        with self.lock():
            try:
                self._execute(
                    f"INSERT INTO {self._table} VALUES (?, ?)",
                    (key, jvalue),
                )
            except sqlite3.IntegrityError:
                self._execute(
                    f"UPDATE {self._table} SET value=? WHERE key=?",
                    (jvalue, key),
                )

    def __delitem__(self, key: str):
        if key in self:
            self._execute(f"DELETE FROM {self._table} WHERE key=?", (key,))
        else:
            raise KeyError

    @contextmanager
    def lock(self):
        if not self._locks:
            self._execute("BEGIN IMMEDIATE TRANSACTION")
        self._locks += 1
        try:
            yield
        finally:
            self._locks -= 1
            if not self._locks:
                self._execute("COMMIT")

    def __getattr__(self, name: str) -> "KV":
        self._attr = name
        return self

    def __iadd__(self, other: int):
        if self._attr is None:  # type: ignore
            raise AttributeError(
                f"'{type(self).__name__}' object has no attribute '_attr'"
            )
        val = self[self._attr]
        val += other
        self[self._attr] = val


if __name__ == "__main__":
    kv = KV(default=0)
    kv["a"] = 1
    kv["b"] = 2
    kv["c"] = 3
    for i in ["a", "b", "c", "d"]:
        print(f"{i}: {kv[i]}")

    e = kv.e
    e += 1
    print(f"e: {kv['e']}")

    kv.f += 1  # type: ignore
    print(f"f: {kv['f']}")



from copy import deepcopy
from click import echo, prompt, Choice as choice, group, argument
from typing import Dict, Hashable, List, TypeVar, Tuple
from result import Result, Ok, Err
from .log_utils import sh, proc


K = TypeVar("K", bound=Hashable)
V = TypeVar("V")


def append(map: Dict[K, List[V]], key: K, value: V) -> None:
    try:
        map[key]
    except KeyError:
        map[key] = []
    finally:
        map[key].append(value)


def version_key(version: str) -> Tuple[int, ...]:
    return tuple(map(int, version.split(".")))


def sort_versions(versions: List[str]) -> List[str]:
    new_versions: List[str] = []

    for version in versions:
        try:
            version_key(version)
            new_versions.append(version)
        except ValueError:
            continue

    new_versions.sort(key=version_key, reverse=True)
    return new_versions


def list_versions() -> None:
    versions: List[str] = sh("pyenv install --list").unwrap_or([""])
    global _all_versions
    filtered: List[str] = []
    for version in versions[1:]:
        if version.startswith("2.") or version.startswith("3."):
            try:
                version_key(version)
                if "dev" not in version and "rc" not in version:
                    filtered.append(version)
            except ValueError:
                continue
    _all_versions = sort_versions(filtered)


def newest_version():
    versions = sh("pyenv versions --bare --skip-envs").unwrap_or(["system"])
    numeric = [ve for ve in versions if "/" not in ve and len(ve) > 0]
    try:
        numeric.remove("system")
    except Exception:
        pass
    sort_versions(numeric)
    global _installed
    global _all_versions
    _installed = deepcopy(numeric)
    return numeric[0]


_all_versions: List[str] = []
_installed: List[str] = []

if len(_all_versions) == 0:
    list_versions()

if len(_installed) == 0:
    newest = newest_version()


def not_installed(version: str) -> Result[str, str]:
    echo(f"python version {version} is not installed")
    response = prompt(
        f"python version {version} is not installed. install it?",
        type=choice(["y", "n"]),
        default="y",
    )
    if response == "y":
        install = proc(f"pyenv install {version}")
        match install:
            case Ok(_):
                echo(f"installed python version {version}")
                return Ok("accept_ok")
            case Err(e):
                echo(e)
                return Err("accept_err")
    elif response == "n":
        return Err("decline_err")
    return Err("unreachable")


@group(name="utilki")
def cli():
    pass


@cli.command()
@argument(
    "version",
    type=choice(_all_versions),
    default=_installed[0],
)
@argument(
    "venv",
    type=str,
    default="",
)
def create(version: str, venv: str = ""):
    """
    ut create [version] [venv]
    """
    global _all_versions
    global _installed

    if version == _installed[0]:
        echo(f"creating venv with default version {version}")
    else:
        echo(f"creating venv with python version {version}")
    if version not in _installed:
        install = not_installed(version)
        match install:
            case Ok(_):
                pass
            case Err("accept_err"):
                return
            case Err("decline_err"):
                return
            case _:
                pass

    elif version not in _all_versions:
        print("lmao")
        echo(f"python version {version} is not installed")

    venv = venv if venv else prompt("enter venv name", type=str)
    create = sh(f"pyenv virtualenv {version} {venv}")

    match create:
        case Ok(_):
            echo(f"created venv {venv} with python version {version}")
        case Err(e):
            echo(e)
            echo("failed to create new venv! 😭")


@cli.command()
@argument(
    "venv",
    type=str,
    default="",
)
def delete(venv: str = ""):
    """
    ut delete [venv]
    """
    venv = venv if venv else prompt("enter venv name", type=str)
    delete_ = proc(f"pyenv virtualenv-delete -f {venv}")

    match delete_:
        case Ok(_):
            echo(f"deleted venv {venv}")
        case Err(e):
            echo(e)
            echo("failed to delete venv! 😭")


@cli.command()
def v():
    """
    ut v
    """
    versions = sh("pyenv versions --bare --skip-aliases").unwrap_or([""])
    echo("\n".join(versions))


if __name__ == "__main__":
    cli()



import ast
import json
import os
from dataclasses import Field
from datetime import date, datetime
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    Iterable,
    List,
    Tuple,
    TypeVar,
    Union,
)

from pydantic import Field as ModelField
from pydantic.fields import FieldInfo

Defaults = Union[
    datetime,
    bool,
    int,
    float,
    str,
    List[bool],
    List[int],
    List[float],
    List[str],
]


singles = [int, float, str, bool, datetime]
lists = [
    List[int],
    List[float],
    List[str],
    List[bool],
    List[datetime],
    List[List[int]],
    List[List[float]],
    List[List[str]],
    List[List[bool]],
]
options = [
    Union[int, None],
    Union[float, None],
    Union[str, None],
    Union[bool, None],
    Union[datetime, None],
    Union[List[int], None],
    Union[List[float], None],
    Union[List[str], None],
    Union[List[bool], None],
    Union[List[datetime], None],
]

dicts = (
    [Dict[str, Any], Dict[str, str], Dict[str, int], Dict[str, float]]
    + [
        Dict[str, Dict[str, Any]],
        Dict[str, Dict[str, str]],
        Dict[str, Dict[str, int]],
        Dict[str, Dict[str, float]],
    ]
    + [
        Dict[Any, Any],
        Dict[Any, str],
        Dict[Any, int],
        Dict[Any, float],
        Dict[Any, bool],
    ]
)

types_we_support = singles + lists + options + dicts


defaults = Union[
    int,
    float,
    str,
    bool,
    datetime,
    List[int],
    List[float],
    List[str],
    List[bool],
]

IsDefault = bool


class ParseError(Exception):
    pass


class TaskMixin:
    __dataclass_fields__: ClassVar[Dict[str, Field]]  # type: ignore
    # model_fields: ClassVar[Dict[str, FieldInfo]]

    @classmethod
    def __init__(cls, **kwargs: Any): ...

    @classmethod
    def create(cls):
        params: List[Tuple[str, Any]] = []  # type: ignore
        if hasattr(cls, "__dataclass_fields__"):
            fields: Iterable[Field[Any]] = cls.__dataclass_fields__.values()  # type: ignore # noqa
            for field in fields:
                params.append((field.name, field.type))
        elif hasattr(cls, "model_fields"):
            for field_name, model_field in cls.model_fields.items():  # type: ignore # noqa
                params.append((field_name, model_field.annotation))  # type: ignore # noqa
        else:
            raise TypeError("Invalid type")
        task = cls(**{name: cls.parse(name, type) for name, type in params})
        return task

    @classmethod
    def get_default(cls, name_: str) -> Tuple[IsDefault, Defaults]:
        env_var = os.environ.pop(name_, None)
        if env_var is not None:
            return False, env_var
        if hasattr(cls, "__dataclass_fields__"):
            result = cls.__dataclass_fields__[name_].default
            return True, result
        elif hasattr(cls, "model_fields"):
            model_field: FieldInfo = cls.model_fields[name_]  # type: ignore # noqa
            default_model = model_field.get_default(call_default_factory=True)  # type: ignore # noqa
            return True, default_model  # type: ignore
        else:
            raise TypeError("Invalid type")

    @classmethod
    def parse(cls, name_: str, type_: type):
        is_default, value = cls.get_default(name_)
        if is_default:
            if type_ not in types_we_support:
                raise TypeError("Invalid type")
            if type_ != type(value) and type_ in singles:
                raise TypeError("Invalid type")
            if type_ in lists and not isinstance(value, list):
                raise TypeError("Invalid type")
            if type_ in dicts and not isinstance(value, dict):
                raise TypeError("Invalid type")
            else:
                return value
        else:
            try:
                res = parse_variations(type_, value, name_)
            except Exception as e:
                raise ParseError(f"{name_=} {value=} {type_=})") from e
            return res

    def update(self, param_dict: Dict[str, Any]):
        for param, value in param_dict.items():
            if hasattr(self, param):
                setattr(self, param, value)
            else:
                raise ValueError(f"Invalid parameter {param}")


def parse_bool(param: str) -> bool:
    if param in {"True", "true", True}:
        return True
    elif param in {"False", "false", False}:
        return False
    else:
        msg = "Invalid boolean format"
        raise TypeError(msg)


T = TypeVar("T")


def parse_options(value: str, type_: Callable[[str], T]) -> Union[T, None]:
    if value in ["None", "none", None, "null", "NULL", ""]:
        return None
    else:
        return type_(value)


def parse_list(
    list_str: str,
    type_: Callable[[str], T],
    name_: str,
) -> List[T]:
    if list_str.startswith("[") and list_str.endswith("]"):
        try:
            return json.loads(list_str)
        except json.JSONDecodeError:
            return ast.literal_eval(list_str)
    elif "," in list_str:
        return list(map(type_, list_str.split(",")))
    else:
        raise ValueError(f"Invalid list format {name_}: {list_str}")


def get_date(param: str):
    n = [int(n) for n in param.split("-")]
    if len(n) == 3:
        return datetime(n[0], n[1], n[2])
    elif len(n) == 6:
        return datetime(n[0], n[1], n[2], n[3], n[4], n[5])
    else:
        raise ValueError("Invalid datetime format")


def parse_variations(type_: type, value: Any, name_: str):
    if type_ == bool:  # noqa: E721
        return parse_bool(value)
    elif type_ == int:  # noqa: E721
        return int(value)
    elif type_ == float:  # noqa: E721
        return float(value)
    elif type_ == List[int]:
        return parse_list(value, int, name_)
    elif type_ == List[str]:
        return parse_list(value, str, name_)
    elif type_ == List[float]:
        return parse_list(value, float, name_)
    elif type_ == List[bool]:
        return parse_list(value, parse_bool, name_)
    elif type_ == Union[int, None]:
        return parse_options(value, int)
    elif type_ == Union[str, None]:
        return parse_options(value, str)
    elif type_ == Union[float, None]:
        return parse_options(value, float)
    elif type_ == Union[bool, None]:
        return parse_options(value, parse_bool)
    elif type_ == Dict[int, Any]:
        return json.loads(value)
    elif type_ == Dict[str, Any]:
        return json.loads(value)
    elif type_ == Dict[float, Any]:
        return json.loads(value)
    elif type_ == Dict[bool, Any]:
        return json.loads(value)
    elif type_ == str:  # noqa: E721
        try:
            res = json.loads(value)
            if not isinstance(res, str):
                return str(res)
            else:
                return res
        except Exception:
            return str(value)
    elif type_ in [datetime, date]:
        # FIXME actually like use proper parsing lmao
        if value.startswith('"') and value.endswith('"'):
            value = value[1:-1]
        has_t_or_colon = ":" in value or "T" in value
        num_parts = len(value.split("-"))
        if has_t_or_colon:
            value = value.replace(" ", "-")
            value = value.replace("T", "-")
            value = value.replace(":", "-")
            return get_date(value)
        elif num_parts == 3:
            return get_date(value)
        else:
            raise TypeError("Invalid datetime format")
    elif type_ in types_we_support:
        try:
            return json.loads(value)
        except Exception:
            try:
                return ast.literal_eval(value)
            except Exception:
                raise TypeError("Invalid type")
